#include <stdlib.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#define USB_DEBUGGING
#ifdef USB_DEBUGGING
# include "usb_debug_only.h"
# include "print.h"
#endif

#include "bool.h"
#include "buttons.h"

/*
This checks the beat lengths to make sure they match the key. The key is an
array that contains how long each beat should be relative to each other. For
example, the key { 2, 3, 1 } represents a double-length beat, a
triple-length beat, and a single-length beat to finish off. 
IMPORTANT: A beat is finished and measured *WHEN THE NEXT BEAT IS BEGUN*. I.e., 
even though the example key contains three elements, I would actually be 
pressing the code key FOUR times, as follows:

Button Pressed     Comment
START              Begin code input
CODE (1)           Begin measuring first beat. key = { }
CODE (2)           End of first beat; store beat time to inputCode buffer,
                      begin measuring second beat. inputCode = { b1 }
CODE (3)           End of second beat; store beat time to inputCode buffer,
                      begin measuring third beat. inputCode = { b1 b2 }
CODE (4)           End of third beat; store beat time to inputCode buffer,
                      begin measuring fourth beat. inputCode = { b1 b2 b3 }
START              Submit inputCode buffer to analyzeCode(). The length of the
                      fourth beat is ignored and not stored to inputCode.
                      
Musically speaking, to match the example key here, you could think of the 
example key as a quarter note, a dotted quarter, an eighth, and another quarter
note. However, since the last beat is not measured, that beat can be of any 
length, provided it is there.

As long as the beats are in the correct proportion to each other, the door will
unlock. The user can set any tempo he likes, as long he doesn't speed up or slow
down too much. 

I have set an accuracy tolerance of +/-30% (float toleranceRatio). This was good
enough to tell the difference between 
{ 3, 3, 3, 2, 1, 3, 2, 1 }  (Star Wars Empire March)
and
{ 3, 3, 3, 2, 1, 2, 1, 2 }  (Apple's "Bad News" voice).

You can decrease toleranceRatio for more security, if you happen to be a
metronome. 
*/

const float toleranceRatio = 0.35; 

// length of your key; change as necessary
#define keyLength 11
const int key[keyLength] = { 2, 2, 2, 3, 1, 2,  2, 2, 2, 3, 1 };
// "My country 'tis of thee, sweet land of liberty"

// maximum length of inputCode buffer. increase this if necessary.
#define inputCodeLength 16 

volatile long timer = 0;
volatile long inputCode[inputCodeLength];
volatile int inputCodeIndex = 0;

typedef enum {
    NO_STATUS = 0x0,
    CODE_SW_UP = 0x1,
    CODE_SW_DOWN = 0x2,
    START_SW_UP = 0x4,
    START_SW_DOWN = 0x8,
    PASSCODE_FINISH = 0x10,
} status_t;
volatile status_t status;

#define LED_CONFIG	(DDRD |= (1<<6))
#define LED_OFF		(PORTD &= ~(1<<6))
#define LED_ON		(PORTD |= (1<<6))
#define CPU_PRESCALE(n)	(CLKPR = 0x80, CLKPR = (n))

void lock(void) 
{ 
#ifdef USB_DEBUGGING
    print("## LOCKING ##\n");
#endif
    LED_OFF; 
    PORTD &= ~(1<<7); 
}
void unlock(void) 
{ 
#ifdef USB_DEBUGGING
    print("## UNLOCKING ##\n");
#endif
    LED_ON; 
    PORTD |= (1<<7); 
}

void init(void)
{
    CPU_PRESCALE(1); // 8 MHz clock

    LED_CONFIG; LED_OFF;
    
    DDRD |= (1<<7); // door strike
    PORTD &= ~(1<<7);

    // init usb for debugging messages
    usb_init(); // can anyone figure out how to get rid of this line safely?
                // I can't get to code input mode without it.

    PORTD |= 0x3; // hold high; active-low buttons
    _delay_ms(100);    
    initButtonInterrupts(); // enable interrupts on INT0 and INT1
}

// INTERRUPT FUNCTIONS
// The following three functions, flagStartDown(), flagCodeDown(), and 
// addCodePulse(), are called by an interrupt handler in buttons.c whenever
// a button is pressed. The interrupt handler checks to see whether the
// appropriate function pointer is set, and invokes it if it is. 
void flagStartDown(void) // sets status to indicate the start button is pressed 
{
    status |= START_SW_DOWN; 
}
void flagCodeDown(void) // sets status to indicate the code button is pressed
{
    status |= CODE_SW_DOWN; 
}

void addCodePulse(void) // add new beat to the inputted code array
{
    if (timer < 150) { // ignore switch bounce, continue timing
        return;
    }
#ifdef USB_DEBUGGING
    print("addCodePulse(): 0x"); phex32(timer); print("\n");
#endif
    inputCode[inputCodeIndex++] = timer; // store beat length to inputCode
    timer = 0;
    if (inputCodeIndex >= inputCodeLength) { // is the input buffer full?
#ifdef USB_DEBUGGING
        print("inputCodeIndex: 0x"); phex(inputCodeIndex);
        print("\n## PASSCODE_FINISH set\n");
#endif
        status |= PASSCODE_FINISH; // if so, force code analysis
        return;
    }
}

// MODE-SETTING FUNCTIONS
// The following four functions, setStandbyMode(), setWaitCodeMode(), 
// setCodeInputMode(), and setCodeAnalyzeMode(), set the status lights
// and interrupt function pointers. They also clear the status flag register
// and reset the timer as necessary.
void setStandbyMode(void) // wait for user to press the start button
{ 
#ifdef USB_DEBUGGING
    print("setStandbyMode()\n");
#endif
    lock();
    timer = 0;
    LED_OFF;
    status = NO_STATUS;
    fp_startSWDown = flagStartDown; 
    fp_codeSWDown = NULL;
}
void setWaitCodeMode(void) // wait for user to begin code input
{ 
#ifdef USB_DEBUGGING
    print("setWaitCodeMode()\n");
#endif
    lock();
    timer = 0;
    int i;
    for (i=0; i<inputCodeLength; i++) {
        inputCode[i] = -1;
    }
    inputCodeIndex = 0;
    LED_OFF;
    status = NO_STATUS;
    fp_startSWDown = NULL;
    fp_codeSWDown = flagCodeDown;
    _delay_ms(50); // make sure the start button isn't mistaken for end-of-code
}
void setCodeInputMode(void) // code input has begun! store beat lengths to
{                           // inputCode buffer, and stop after the start
#ifdef USB_DEBUGGING        // button is pressed.
    print("setCodeInputMode()\n");
#endif
    lock();
    timer = 0;      
    LED_OFF;
    status = NO_STATUS;
    fp_startSWDown = flagStartDown;    
    fp_codeSWDown = addCodePulse; 
}
void setCodeAnalyzeMode(void) // analyzing code. ignore any input.
{
#ifdef USB_DEBUGGING
    print("setCodeAnalyzeMode()\n");
#endif
    lock();
    timer = 0;
    LED_OFF;
    status = NO_STATUS;
    fp_startSWDown = NULL;
    fp_codeSWDown = NULL;
}

// I've commented out lots of stuff here for security. It wouldn't be good
// if a hacker connected his computer to your lock and got the plaintext code
// from it. On the other hand, the microcontroller really shouldn't be
// accessible to him, anyway...
bool analyzeCode(void)
{
    bool result = TRUE;
#ifdef USB_DEBUGGING
//    print("inside analyzeCode()...\n");
#endif
    if (keyLength != inputCodeIndex) {
#ifdef USB_DEBUGGING
//        print("incorrect code length. Should be 0x"); phex16(keyLength);
//        print(" but is 0x"); phex16(inputCodeIndex); print("\n");
#endif
        result = FALSE;
    }
    
    int i, totalKeyLength=0;
    long totalInputKeyLength=0;
    for (i=0; i<keyLength; i++) {
        totalKeyLength += key[i];
        totalInputKeyLength += inputCode[i];
    }
    long avgBeatLength = totalInputKeyLength / totalKeyLength;
    long unitTolerance = (long) ((float) avgBeatLength * toleranceRatio);
#ifdef USB_DEBUGGING
//    print("avgBeatLength: 0x"); phex32(avgBeatLength); print("\n");
//    print("unitTolerance: 0x"); phex32(unitTolerance); print("\n");
#endif    
    
    long thisBeatLength;
    for (i=0; i<keyLength; i++) {
        thisBeatLength = avgBeatLength * key[i];
#ifdef USB_DEBUGGING
/*
        // print info comparing input to code key
        print("0x"); phex16(i);
        print(": 0x"); phex32(inputCode[i]);

        print("   key: 0x"); phex16(key[i]); 
        print(" : "); phex32(thisBeatLength);
        print(" : ");
*/
#endif
        if ( (inputCode[i] > thisBeatLength - unitTolerance) &&
             (inputCode[i] < thisBeatLength + unitTolerance) ) 
        {
#ifdef USB_DEBUGGING
//            print("GOOD\n");
#endif
        } else {
#ifdef USB_DEBUGGING
//            print("BAD\n");
#endif
            result = FALSE;
        }
    }
    return result;
}

int main(void)
{
    init();
    
    while (1) {
standby:
        setStandbyMode();
        while (~status & START_SW_DOWN) ;
        setWaitCodeMode();
        while (~status & CODE_SW_DOWN) {
            _delay_ms(1);
            timer++;
            if (timer > 30000) 
                goto standby;
            if (timer % 1000 < 500)
                LED_ON;
            else
                LED_OFF;
        }    

        setCodeInputMode();
#ifdef USB_DEBUGGING
        print("Timing pulses...\n");
#endif
        while ( (~status & START_SW_DOWN) &&
                (~status & PASSCODE_FINISH) ) {
            _delay_ms(1);
            timer++;
            if (timer > 30000) 
                goto standby;
            if (timer % 500 < 250)
                LED_ON;
            else
                LED_OFF;
        }

        setCodeAnalyzeMode();        

#ifdef USB_DEBUGGING        
        print("Analyzing code...\n");
#endif
        if (analyzeCode()) {
#ifdef USB_DEBUGGING
    print("Good code.\n");
#endif
            unlock();
            _delay_ms(8000);
            lock();
            _delay_ms(100); // make sure noise from relay doesn't trigger button
        } else {
#ifdef USB_DEBUGGING
    print("Bad code.\n");
#endif
            timer = 0;
            while (timer < 4000) {
                if (timer % 125 < 67) {
                    LED_OFF;
                } else {
                    LED_ON;
                }
                _delay_ms(1);
                timer++;
            }
        }
    }
}
